# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Module for CSV playback poll plugin using pandas """

import copy
import logging
import os
import pandas as pd
from threading import Event
from threading import Thread, Condition
import datetime
import time
import async_ingest
from foglamp.common import logger
from foglamp.plugins.common import utils
import json

__author__ = "Rajesh Kumar , Deepanshu Yadav"
__copyright__ = "Copyright (c) 2019 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Test Sample',
        'type': 'string',
        'default': 'csvplayback',
        'readonly': 'true'
    },
    'assetName': {
        'description': 'Name of Asset',
        'type': 'string',
        'default': 'sample',
        'displayName': 'Asset Name',
        'order': '1'
    },
    'csvFilename': {
        'description': 'CSV File name',
        'type': 'string',
        'default': 'some.csv',
        'displayName': 'CSV file name with extension',
        'order': '2'
    },
    'headerRow': {
        'description': 'If header row is preset as first row of CSV file',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Header Row',
        'order': '3'
    },
    'fieldNames': {
        'description': 'Comma separated list of column names, mandatory if headerRow is false',
        'type': 'string',
        'default': 'None',
        'displayName': 'Header columns',
        'order': '4'
    },
    'readingCols': {
        'description': 'Cherry pick data columns with the same/new name',
        'type': 'JSON',
        'default': '{}',
        'displayName': 'Cherry pick column with same/new name',
        'order': '5'
    },
    'historicTimestamps': {
        'description': 'Use original timestamps rather than ingest in real time',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Historic timestamps',
        'order': '6'
    },
    'timestampDeltaFromFile': {
        'description': 'Take the difference of consecutive readings from file and insert'
                       ' these differences into user_ts',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'timestamp Delta From File',
        'order': '7'
    },
    'timestampCol': {
        'description': 'Timestamp header column, mandatory if timestampFromFile is true',
        'type': 'string',
        'default': 'ts',
        'displayName': 'Timestamp column name',
        'order': '8'
    },
    'timestampFormat': {
        'description': 'Timestamp format in File, mandatory if timestampFromFile is true',
        'type': 'string',
        'default': '%Y-%m-%d %H:%M:%S.%f',
        'displayName': 'Timestamp format',
        'order': '9'
    },
    'ingestMode': {
        'description': 'Mode of data ingest - burst/batch',
        'type': 'enumeration',
        'default': 'batch',
        'options': ['batch', 'burst'],
        'displayName': 'Ingest mode',
        'order': '10'
    },
    'sampleRate': {
        'description': 'No. of readings per sec',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Sample Rate',
        'minimum': '1',
        'maximum': '1000000',
        'order': '11'
    },
    'burstInterval': {
        'description': 'Time interval between consecutive bursts in milliseconds',
        'type': 'integer',
        'default': '1000',
        'displayName': 'Burst Interval (ms)',
        'minimum': '1',
        'order': '12'
    },
    'burstSize': {
        'description': 'No. of data points in one burst',
        'type': 'integer',
        'default': '1',
        'displayName': 'Burst size',
        'minimum': '1',
        'order': '13'
    },
    'repeatLoop': {
        'description': 'Read CSV in a loop i.e. on reaching EOF, again go back to beginning of the file',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Read file in a loop',
        'order': '14'
    },
}

# GLOBAL VARIABLES DECLARATION
_FOGLAMP_ROOT = os.getenv("FOGLAMP_ROOT", default='/usr/local/foglamp')
_FOGLAMP_DATA = os.path.expanduser(_FOGLAMP_ROOT + '/data')
_LOGGER = logger.setup(__name__, level=logging.INFO)
producer = None  # A Producer object to read data from csv file
consumer = None  # A Consumer object to ingest data to database
condition = None  # A conditional variable for mutual exclusivity among consumer and producer
wait_event = Event()  # A variable to control the rate of ingest into database and indicate shutdown
c_callback = None
c_ingest_ref = None
_sentinel = object()  # Indicates the file has been read
readingsQueue = None  # Queue/ buffer to store readings.


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        'name': 'CSV Playback',
        'version': '1.8.1',
        'mode': 'async',
        'type': 'south',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the South plugin configuration category
    Returns:
        handle: JSON object to be used in future calls to the plugin
    Raises:
    """
    handle = copy.deepcopy(config)

    try:
        errors = False
        csv_file_name = "{}/{}".format(_FOGLAMP_DATA, handle['csvFilename']['value'])
        if not handle['csvFilename']['value']:
            _LOGGER.exception("csv filename cannot be empty")
            errors = True
        if not os.path.isfile(csv_file_name):
            _LOGGER.exception('csv filename "{}" not found'.format(csv_file_name))
            errors = True
        if int(handle['sampleRate']['value']) < 1 or int(handle['sampleRate']['value']) > 1000000:
            _LOGGER.exception("sampleRate should be in range 1-1000000")
            errors = True
        if int(handle['burstSize']['value']) < 1:
            _LOGGER.exception("burstSize should not be less than 1")
            errors = True
        if int(handle['burstInterval']['value']) < 1:
            _LOGGER.exception("burstInterval should not be less than 1")
            errors = True
        if handle['ingestMode']['value'] not in ['burst', 'batch']:
            _LOGGER.exception("ingestMode should be one of ('burst', 'batch')")
            errors = True
        if handle['historicTimestamps']['value'] == 'true' and \
                (handle['timestampCol']['value'] is None or handle['timestampFormat']['value'] is None):
            _LOGGER.exception("timestamp Column (of csv File) and timestamp Format  must be specified ")
            errors = True
        if handle['timestampDeltaFromFile']['value'] == 'true' and \
                (handle['timestampCol']['value'] is None or handle['timestampFormat']['value'] is None):
            _LOGGER.exception("timestamp Column (of csv File) and timestamp Format  "
                              "must be specified for Delta calculation")
            errors = True
        if handle['historicTimestamps']['value'] == 'true' and \
                (handle['timestampDeltaFromFile']['value'] == 'true'):
            _LOGGER.exception("Both historicTimestamps and  "
                              "timestampDeltaFromFile cannot be True at the same time")
            errors = True
        if errors:
            raise RuntimeError
    except KeyError:
        raise
    except ValueError:
        raise
    except RuntimeError:
        raise
    else:
        return handle


def plugin_start(handle):
    """ Extracts data from the CSV and returns it in a JSON document as a Python dict.
    Available for async mode only.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        a playback reading in a JSON document, as a Python dict
    """
    global producer, consumer, condition, readingsQueue, wait_event
    readingsQueue = []
    condition = Condition()
    producer = Producer(handle)
    consumer = Consumer(handle)

    # Clearing the wait event , In case of reconfigure , Shutdown is called . During shutdown the wait_event
    # is set. So if plugin starts again we need to clear this event once again.
    wait_event.clear()
    producer.start()
    consumer.start()


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.info("Old config for playback plugin {} \n new config {}".format(handle, new_config))
    plugin_shutdown(handle)
    new_handle = plugin_init(new_config)
    plugin_start(new_handle)
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup, to be called prior to the South plugin service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _LOGGER.info('csv playback Plugin Shutting down')

    global producer, consumer, wait_event, condition, readingsQueue

    # The wait event flag needs to be set to shut down the plugin
    wait_event.set()
    time.sleep(2)  # It is done to allow the consumer thread to figure out that wait_event flag has been set.

    if producer is not None:
        producer._tstate_lock = None
        producer = None
    if consumer is not None:
        consumer._tstate_lock = None
        consumer = None
    if readingsQueue:
        readingsQueue = []
    condition = None

    _LOGGER.info('csv playback Plugin Shut down.')


def plugin_register_ingest(handle, callback, ingest_ref):
    """Required plugin interface component to communicate to South C server

    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C->ingest method
        ingest_ref: C opaque object required to passed back to C->ingest method
    """
    global c_callback, c_ingest_ref
    c_callback = callback
    c_ingest_ref = ingest_ref


class Producer(Thread):
    def __init__(self, handle):
        """
        Initializes the Producer class to read readings from csv file.
        Args:
            handle: The configuration of the plugin
        """
        super(Producer, self).__init__()
        self.handle = handle
        self.prv_readings_ts = None
        self.timestamp_interval = None
        self.max_capacity = 3  # The maximum capacity of the queue

        try:
            if self.handle['ingestMode']['value'] == 'burst':
                burst_interval = int(self.handle['burstInterval']['value'])
                self.period = round(burst_interval / 1000.0, len(str(burst_interval)) + 1)
            else:
                recs = int(self.handle['sampleRate']['value'])
                self.period = round(1.0 / recs, len(str(recs)) + 1)

        except ZeroDivisionError:
            _LOGGER.warning('sampleRate must be greater than 0, defaulting to 1')
            self.period = 1.0

    def read_csv_file(self):
        """Creates an iterator to retrieve the data from csv file
        Returns: Returns an iterator in which each element contains a small Data frame

        """
        csv_path = '{}/{}'.format(_FOGLAMP_DATA, self.handle['csvFilename']['value'])
        reading_cols = self.handle['readingCols']['value']
        header_rows = True if self.handle['headerRow']['value'] == 'true' else False

        if not header_rows:
            data = pd.read_csv(csv_path, iterator=True, chunksize=int(self.handle['sampleRate']['value']),
                               names=self.handle['fieldNames']['value'].split(','))
        else:
            data = pd.read_csv(csv_path, iterator=True, chunksize=int(self.handle['sampleRate']['value']))

        if self.handle['ingestMode']['value'] == 'burst':
            if not header_rows:
                data = pd.read_csv(csv_path, iterator=True, chunksize=int(self.handle['burstSize']['value']),
                                   names=self.handle['fieldNames']['value'].split(','))
            else:
                data = pd.read_csv(csv_path, iterator=True, chunksize=int(self.handle['burstSize']['value']))

        if isinstance(reading_cols, dict) and reading_cols:
            data.rename(columns=reading_cols, inplace=True)
        if isinstance(reading_cols, str) and not reading_cols == '{}':
            reading_cols = json.loads(reading_cols)
            data.rename(columns=reading_cols, inplace=True)

        return data

    def run(self):
        reading_cols = self.handle['readingCols']['value']
        data = self.read_csv_file()

        global readingsQueue
        eof = True

        while eof:

            # If plugin shutdown has been called
            global wait_event
            if wait_event.is_set():
                return

            # Do not exceed maximum capacity of Queue
            condition.acquire()
            if len(readingsQueue) >= self.max_capacity:
                condition.wait()

            data_dict = {}
            if self.handle['ingestMode']['value'] == 'burst':
                try:
                    element = next(data)
                    data_dict.update({'data': element.to_dict('records')})
                    readingsQueue.append(data_dict)
                except StopIteration:
                    # Check if we want to play the file again
                    eof = True if self.handle['repeatLoop']['value'] == 'true' else False
                    if not eof:
                        readingsQueue.append(_sentinel)
                    else:
                        data = self.read_csv_file()

            else:
                try:

                    element = next(data)
                    # Handle the data type of reading_cols differently.
                    if isinstance(reading_cols, dict) and reading_cols:
                        element = element.loc[:, list(reading_cols.values())]
                    if isinstance(reading_cols, str) and not reading_cols == '{}':
                        reading_cols = json.loads(reading_cols)
                        element = element.loc[:, list(reading_cols.values())]

                    readingsQueue.append(element)

                except StopIteration:
                    eof = True if self.handle['repeatLoop']['value'] == 'true' else False
                    if not eof:
                        readingsQueue.append(_sentinel)
                    else:
                        data = self.read_csv_file()

            # Notifying the consumer that data has been read.
            condition.notify()
            condition.release()


class Consumer(Thread):
    def __init__(self, handle):
        super(Consumer, self).__init__()
        self.handle = handle
        try:
            if self.handle['ingestMode']['value'] == 'burst':
                burst_interval = int(self.handle['burstInterval']['value'])
                self.period = round(burst_interval / 1000.0, len(str(burst_interval)) + 1)
            else:
                self.period = 1

        except ZeroDivisionError:
            _LOGGER.warning('sampleRate must be greater than 0, defaulting to 1')
            self.period = 1.0

    def run(self):
        global readingsQueue

        # Initialize the current time required for time stamp delta if required
        c = datetime.datetime.now(datetime.timezone.utc).astimezone()
        ts_diff = None

        while True:
            # Check if shutdown is called
            global wait_event
            if wait_event.is_set():
                return

            # Check if queue is empty
            condition.acquire()
            if len(readingsQueue) == 0:
                condition.wait()
            start_time = datetime.datetime.now()

            if self.handle['ingestMode']['value'] == 'burst':
                reads = readingsQueue.pop(0)
                data = pd.DataFrame(reads['data'])
            else:
                data = readingsQueue.pop(0)

                if data is _sentinel:
                    readingsQueue.append(_sentinel)
                    break
                reads = data.to_dict('records')

            # Modifying the time stamps if a column that has time stamps
            is_timestamp_col = False

            if self.handle['historicTimestamps']['value'] == 'true':
                ts_col = self.handle['timestampCol']['value']
                try:
                    org_time_col = data[ts_col]
                except KeyError:
                    _LOGGER.exception('The column name does not exist in the file.')

                ts_format = self.handle['timestampFormat']['value']
                timestamp = pd.to_datetime(org_time_col, format=ts_format).array
                is_timestamp_col = True

            elif self.handle['timestampDeltaFromFile']['value'] == 'true':
                ts_col = self.handle['timestampCol']['value']
                try:
                    org_time_col = data[ts_col]
                except KeyError:
                    _LOGGER.exception('The column name does not exist in the file.')
                ts_format = self.handle['timestampFormat']['value']
                ts_array = pd.to_datetime(org_time_col, format=ts_format).array
                timestamp = []

                # Calculate time difference .This will be added to all readings except the first.
                if ts_diff is None:
                    ts_diff = ts_array[1] - ts_array[0]

                for k in range(len(ts_array)):
                    timestamp.append(c)
                    c = c + ts_diff
                is_timestamp_col = True

            if self.handle['ingestMode']['value'] == 'batch':

                original_timestamp = datetime.datetime.now(datetime.timezone.utc).astimezone()
                uniform_interval = int((1.0/len(reads))*1000000)
                rows_in_data_frame = len(reads)

                for j in range(rows_in_data_frame):

                    if is_timestamp_col:
                        modified_timestamp = str(timestamp[j])
                    else:
                        modified_timestamp = str(original_timestamp.replace(microsecond=uniform_interval*j))
                    readings = {
                        'asset': self.handle['assetName']['value'],
                        'timestamp': modified_timestamp,
                        'readings':   reads[j]
                    }

                    # Ingesting into database
                    async_ingest.ingest_callback(c_callback, c_ingest_ref, readings)

            else:
                # don't differ time stamps in case of burst mode
                timestamp = utils.local_timestamp()
                readings = {
                    'asset': self.handle['assetName']['value'],
                    'timestamp': timestamp,
                    'readings': reads
                }

                # Ingesting into database
                async_ingest.ingest_callback(c_callback, c_ingest_ref, readings)

            end_time = datetime.datetime.now()

            # Notifying the Producer
            condition.notify()
            condition.release()

            # Wait for a fixed interval of time
            wait_event.wait(timeout=self.period - (end_time - start_time).total_seconds())
